//-----------------------------------------------------------------------------
// vflt_16.h : LPF(Non-SIMD)
// Copyright (c) 2005-2008  N099
//-----------------------------------------------------------------------------
#include "vflt.h"

// 3tap LPF
NOA static void lpf_v_3(pix * restrict dst, const pix * restrict src,
                        int h, int w, int s)
{
    int i,j;
    pint16c src0, src1, src2;
    pint16r dst0;
    int tmp;

    w *= (sizeof(pix) / sizeof(int16_t));
    
    for(i=0; i<h; i++){
        dst0 = (pint16r)dst;
        src1 = (pint16c)src;
        src0 = (pint16c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src2 = (pint16c)(src + s * ((i+1) >= h ? -1 : 1));
        
        for(j=0; j<w; ++j){
            tmp  = src0[j] + src2[j];
            tmp += src1[j] * 2 + 2;
            tmp >>= 2;
            *dst0++ = (int16_t)tmp;
        }
        
        dst += s;
        src += s;
    }
}


// 5tap LPF
NOA static void lpf_v_5(pix * restrict dst, const pix * restrict src,
                        int h, int w, int s)
{
    int i,j;
    pint16c src0, src1, src2, src3, src4;
    pint16r dst0;
    int tmp;

    w *= (sizeof(pix) / sizeof(int16_t));
  
    for(i=0; i<h; i++){
        dst0 = (pint16r)dst;
        src2 = (pint16c)src;
        src1 = (pint16c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src3 = (pint16c)(src + s * ((i+1) >= h ? -1 : 1));
        src0 = (pint16c)(src + s * ((i-2) >= 0 ? -2 : 2 - 2*i));
        src4 = (pint16c)(src + s * ((i+2) >= h ? 2*(h-1-i)-2 : 2));
        
        for(j=0; j<w; ++j){
            tmp  = (*(src0+j) + *(src4+j)) * -2606;
            tmp += (*(src1+j) + *(src3+j)) *  8192;
            tmp += *(src2+j) * 21596 + (1<<14) - 1;
            *dst0++ = (int16_t)(tmp >> 15);
        }
        
        dst += s;
        src += s;
    }
}


// 7tap LPF
NOA static void lpf_v_7(pix * restrict dst, const pix * restrict src,
                        int h, int w, int s)
{
    int i,j;
    pint16c src0, src1, src2, src3, src4, src5, src6;
    pint16r dst0;
    int tmp;

    w *= (sizeof(pix) / sizeof(int16_t));
    
    for(i=0; i<h; i++){
        dst0 = (pint16r)dst;
        src3 = (pint16c)src;
        src2 = (pint16c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src4 = (pint16c)(src + s * ((i+1) >= h ? -1 : 1));
        src1 = (pint16c)(src + s * ((i-2) >= 0 ? -2 : 2 - 2*i));
        src5 = (pint16c)(src + s * ((i+2) >= h ? 2*(h-1-i)-2 : 2));
        src0 = (pint16c)(src + s * ((i-3) >= 0 ? -3 : 3 - 2*i));
        src6 = (pint16c)(src + s * ((i+3) >= h ? 2*(h-1-i)-3 : 3));
            
        for(j=0; j<w; ++j){
            tmp  = (*(src0+j) + *(src6+j)) *  1315;
            tmp += (*(src1+j) + *(src5+j)) * -4000;
            tmp += (*(src2+j) + *(src4+j)) *  6876;
            tmp += *(src3+j) * 24386 + (1<<14) - 1;
            *dst0++ = (int16_t)(tmp >> 15);
        }
        
        dst += s;
        src += s;
    }
}


// 9tap LPF
NOA static void lpf_v_9(pix * restrict dst, const pix * restrict src,
                        int h, int w, int s)
{
    int i,j;
    pint16c src0, src1, src2, src3, src4, src5, src6, src7, src8;
    pint16r dst0;
    int tmp;

    w *= (sizeof(pix) / sizeof(int16_t));

    for(i=0; i<h; i++){
        dst0 = (pint16r)dst;
        src4 = (pint16c)src;
        src3 = (pint16c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src5 = (pint16c)(src + s * ((i+1) >= h ? -1 : 1));
        src2 = (pint16c)(src + s * ((i-2) >= 0 ? -2 : 2 - 2*i));
        src6 = (pint16c)(src + s * ((i+2) >= h ? 2*(h-1-i)-2 : 2));
        src1 = (pint16c)(src + s * ((i-3) >= 0 ? -3 : 3 - 2*i));
        src7 = (pint16c)(src + s * ((i+3) >= h ? 2*(h-1-i)-3 : 3));
        src0 = (pint16c)(src + s * ((i-4) >= 0 ? -4 : 4 - 2*i));
        src8 = (pint16c)(src + s * ((i+4) >= h ? 2*(h-1-i)-4 : 4));
    
        for(j=0; j<w; ++j){
            tmp  = (*(src0+j) + *(src8+j)) *  -766;
            tmp += (*(src1+j) + *(src7+j)) *  2285;
            tmp += (*(src2+j) + *(src6+j)) * -4237;
            tmp += (*(src3+j) + *(src5+j)) *  5907;
            tmp += *(src4+j) * 26390 + (1<<14) - 1;
            *dst0++ = (int16_t)(tmp >> 15);
        }
        
        dst += s;
        src += s;
    }
}


// 11tap LPF
NOA static void lpf_v_11(pix * restrict dst, const pix * restrict src,
                         int h, int w, int s)
{
    int i,j;
    pint16c src0, src1, src2, src3, src4, src5;
    pint16c src6, src7, src8, src9, srca;
    pint16r dst0;
    int tmp;
    
    w *= (sizeof(pix) / sizeof(int16_t));
    
    for(i=0; i<h; i++){
        dst0 = (pint16r)dst;
        src5 = (pint16c)src;
        src4 = (pint16c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src6 = (pint16c)(src + s * ((i+1) >= h ? -1 : 1));
        src3 = (pint16c)(src + s * ((i-2) >= 0 ? -2 : 2 - 2*i));
        src7 = (pint16c)(src + s * ((i+2) >= h ? 2*(h-1-i)-2 : 2));
        src2 = (pint16c)(src + s * ((i-3) >= 0 ? -3 : 3 - 2*i));
        src8 = (pint16c)(src + s * ((i+3) >= h ? 2*(h-1-i)-3 : 3));
        src1 = (pint16c)(src + s * ((i-4) >= 0 ? -4 : 4 - 2*i));
        src9 = (pint16c)(src + s * ((i+4) >= h ? 2*(h-1-i)-4 : 4));
        src0 = (pint16c)(src + s * ((i-5) >= 0 ? -5 : 5 - 2*i));
        srca = (pint16c)(src + s * ((i+5) >= h ? 2*(h-1-i)-5 : 5));
        
        for(j=0; j<w; ++j){
            tmp  = (*(src0+j) + *(srca+j)) *   518;
            tmp += (*(src1+j) + *(src9+j)) * -1442;
            tmp += (*(src2+j) + *(src8+j)) *  2693;
            tmp += (*(src3+j) + *(src7+j)) * -3993;
            tmp += (*(src4+j) + *(src6+j)) *  4980;
            tmp += *(src5+j) * 27256 + (1<<14) - 1;
            *dst0++ = (int16_t)(tmp >> 15);
        }
        
        dst += s;
        src += s;
    }
}


// 13tap LPF
NOA static void lpf_v_13(pix * restrict dst, const pix * restrict src,
                         int h, int w, int s)
{
    int i,j;
    pint16c src0, src1, src2, src3, src4, src5, src6;
    pint16c src7, src8, src9, srca, srcb, srcc;
    pint16r dst0;
    int tmp;

    w *= (sizeof(pix) / sizeof(int16_t));

    for(i=0; i<h; i++){
        dst0 = (pint16r)dst;
        src6 = (pint16c)src;
        src5 = (pint16c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src7 = (pint16c)(src + s * ((i+1) >= h ? -1 : 1));
        src4 = (pint16c)(src + s * ((i-2) >= 0 ? -2 : 2 - 2*i));
        src8 = (pint16c)(src + s * ((i+2) >= h ? 2*(h-1-i)-2 : 2));
        src3 = (pint16c)(src + s * ((i-3) >= 0 ? -3 : 3 - 2*i));
        src9 = (pint16c)(src + s * ((i+3) >= h ? 2*(h-1-i)-3 : 3));
        src2 = (pint16c)(src + s * ((i-4) >= 0 ? -4 : 4 - 2*i));
        srca = (pint16c)(src + s * ((i+4) >= h ? 2*(h-1-i)-4 : 4));
        src1 = (pint16c)(src + s * ((i-5) >= 0 ? -5 : 5 - 2*i));
        srcb = (pint16c)(src + s * ((i+5) >= h ? 2*(h-1-i)-5 : 5));
        src0 = (pint16c)(src + s * ((i-6) >= 0 ? -6 : 6 - 2*i));
        srcc = (pint16c)(src + s * ((i+6) >= h ? 2*(h-1-i)-6 : 6));
        
        for(j=0; j<w; ++j){
            tmp  = (*(src0+j) + *(srcc+j)) *  -363;
            tmp += (*(src1+j) + *(srcb+j)) *   981;
            tmp += (*(src2+j) + *(srca+j)) * -1841;
            tmp += (*(src3+j) + *(src9+j)) *  2822;
            tmp += (*(src4+j) + *(src8+j)) * -3737;
            tmp += (*(src5+j) + *(src7+j)) *  4389;
            tmp += *(src6+j) * 28266 + (1<<14);
            *dst0++ = (int16_t)(tmp >> 15);
        }
        
        dst += s;
        src += s;
    }
}


// 15tap LPF
NOA static void lpf_v_15(pix * restrict dst, const pix * restrict src,
                         int h, int w, int s)
{
    int i,j;
    pint16c src0, src1, src2, src3, src4, src5, src6, src7;
    pint16c src8, src9, srca, srcb, srcc, srcd, srce;
    pint16r dst0;
    int tmp;

    w *= (sizeof(pix) / sizeof(int16_t));
    
    for(i=0; i<h; i++){
        dst0 = (pint16r)dst;
        src7 = (pint16c)src;
        src6 = (pint16c)(src + s * ((i-1) >= 0 ? -1 : 1));
        src8 = (pint16c)(src + s * ((i+1) >= h ? -1 : 1));
        src5 = (pint16c)(src + s * ((i-2) >= 0 ? -2 : 2 - 2*i));
        src9 = (pint16c)(src + s * ((i+2) >= h ? 2*(h-1-i)-2 : 2));
        src4 = (pint16c)(src + s * ((i-3) >= 0 ? -3 : 3 - 2*i));
        srca = (pint16c)(src + s * ((i+3) >= h ? 2*(h-1-i)-3 : 3));
        src3 = (pint16c)(src + s * ((i-4) >= 0 ? -4 : 4 - 2*i));
        srcb = (pint16c)(src + s * ((i+4) >= h ? 2*(h-1-i)-4 : 4));
        src2 = (pint16c)(src + s * ((i-5) >= 0 ? -5 : 5 - 2*i));
        srcc = (pint16c)(src + s * ((i+5) >= h ? 2*(h-1-i)-5 : 5));
        src1 = (pint16c)(src + s * ((i-6) >= 0 ? -6 : 6 - 2*i));
        srcd = (pint16c)(src + s * ((i+6) >= h ? 2*(h-1-i)-6 : 6));
        src0 = (pint16c)(src + s * ((i-7) >= 0 ? -7 : 7 - 2*i));
        srce = (pint16c)(src + s * ((i+7) >= h ? 2*(h-1-i)-7 : 7));
        
        for(j=0; j<w; ++j){
            tmp  = (*(src0+j) + *(srce+j)) *   265;
            tmp += (*(src1+j) + *(srcd+j)) *  -695;
            tmp += (*(src2+j) + *(srcc+j)) *  1299;
            tmp += (*(src3+j) + *(srcb+j)) * -2021;
            tmp += (*(src4+j) + *(srca+j)) *  2766;
            tmp += (*(src5+j) + *(src9+j)) * -3417;
            tmp += (*(src6+j) + *(src8+j)) *  3862;
            tmp += *(src7+j) * 28650 + (1<<14) - 1;
            *dst0++ = (int16_t)(tmp >> 15);
        }
        
        dst += s;
        src += s;
    }
}


const VLPF vflt_16[] =
{ lpf_v_3, lpf_v_5, lpf_v_7, lpf_v_9, lpf_v_11, lpf_v_13, lpf_v_15, NULL };


